/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <fcntl.h>
#include "unnamed.h"

typedef struct
{
  uint32_t end;
  char label[16];
} country_t;

typedef struct
{
  int last_country_index;
  FILE *country_fp;
  uint32_t ttl;
  int trace;
} module_data_t;

static void term(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(module->country_fp != NULL) (void)fclose(module->country_fp);
  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  struct stat st;
  const char *filename;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;

  module->trace = state->config_get_int(state, "country.trace", state->trace);

  module->ttl = (uint32_t)state->config_get_int(state, "country.ttl", DEFAULT_TTL);

  filename = state->config_get_const(state, "country.db");
  if(filename == NULL)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "country.db not set");
    free(module);
    return -1;
  }

  module->country_fp = fopen(filename, "r");
  if(module->country_fp == NULL)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "Unable to open %s", filename);
    free(module);
    return -1;
  }
  memset(&st, 0, sizeof(st));
  if(fstat(fileno(module->country_fp), &st) != 0)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "Unable to fstat %s", filename);
    (void)fclose(module->country_fp);
    free(module);
    return -1;
  }

  module->last_country_index = (int)st.st_size / (int)sizeof(country_t);

  state->module_data[module_index] = module;
  state->config_free_node(state, "country");
  return 0;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  answer_t ret = ANSWER_UNSATISFIED;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(lazy != 0) return ANSWER_UNSATISFIED;
  if(dclass != CLASS_IN && dclass != CLASS_ANY) return ANSWER_UNSATISFIED;
  if(client->header.flags.ra == 0) return ANSWER_UNSATISFIED;
  if(type == TYPE_PTR)
  {
    address_t a;
    int begin, mid, end;
    rr_t *rr;
    country_t country;
    if(state->is_subdomain("in-addr.arpa", (char*)name->value) != 0)
    {
      if(state->addr_prefix(name, &a) != 4) return ANSWER_UNSATISFIED;
    }
    else if(state->is_subdomain("ip6.arpa", (char*)name->value) != 0)
    {
      if(state->addr6_prefix(name, &a) != 16) return ANSWER_UNSATISFIED;
    }
    else
    {
      return ANSWER_UNSATISFIED;
    }

    if(a.a.sockaddr.sa_family == AF_INET)
    {
      uint32_t ip;
      country.end = 0;
      country.label[0] = '\0';
      begin = 0;
      end = module->last_country_index;
      mid = module->last_country_index / 2;
      ip = ntohl(a.a.sockaddr_in.sin_addr.s_addr);
/*todo: replace with bsearch? */
      while(begin < mid)
      {
        if(fseek(module->country_fp, (long)mid * (long)sizeof(country_t), SEEK_SET) == -1) return ANSWER_UNSATISFIED;
        if(fread(&country, sizeof(country), 1, module->country_fp) != 1) return ANSWER_UNSATISFIED;
        if(country.end < ip)
          begin = mid + 1;
        else
          end = mid;
        mid = (begin + end) / 2;
      }

      if(fseek(module->country_fp, (long)begin * (long)sizeof(country_t), SEEK_SET) == -1) return ANSWER_UNSATISFIED;
      if(fread(&country, sizeof(country), 1, module->country_fp) != 1) return ANSWER_UNSATISFIED;
      if(ip >= country.end)
      {
        mid = end;
        if(fseek(module->country_fp, (off_t)mid * (off_t)sizeof(country_t), SEEK_SET) == -1) return ANSWER_UNSATISFIED;
        if(fread(&country, sizeof(country), 1, module->country_fp) != 1) return ANSWER_UNSATISFIED;
      }

      if((rr = state->rr_create(name, TYPE_PTR, CLASS_IN, module->ttl, RCODE_NO_ERROR)) != NULL)
      {
        uint8_t *ad = (uint8_t*)&a.a.sockaddr_in.sin_addr.s_addr;
        if(state->buffer_set(&rr->p.ptr.name, "%u.%u.%u.%u.%s.unknown", ad[3], ad[2], ad[1], ad[0], country.label) == 0)
        {
          if(state->add_rr(client, rr_section, &rr) == 0)
          {
            INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %u.%u.%u.%u.%s.unknown", (char*)name->value, state->class_name(dclass), state->type_name(type), ad[3], ad[2], ad[1], ad[0], country.label);
            ret = ANSWER_SATISFIED;
          }
        }
        state->free_rr(&rr);
      }
    }
    else
    {
      return ANSWER_UNSATISFIED;
    }
  }
  else if(type == TYPE_A)
  {
    char *unknown = strstr((char*)name->value, ".unknown");
    if(unknown != NULL)
    {
      unsigned long offset = (unsigned long)unknown - (unsigned long)name->value;
      if(name->value[offset + sizeof(".unknown") - 1] == 0)
      {
        address_t a;
        rr_t *rr;
        if((rr = state->rr_create(name, TYPE_A, CLASS_IN, module->ttl, RCODE_NO_ERROR)) == NULL) return ANSWER_UNSATISFIED;
        if(state->addr_prefix(name, &a) == 0)
        {
          state->free_rr(&rr);
          return ANSWER_UNSATISFIED;
        }
        memcpy(&rr->p.address, &a, sizeof(rr->p.address));
        if(state->add_rr(client, rr_section, &rr) == 0)
        {
          char address[IPCHAR_SIZE];
          address[0] = '\0';
          INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)name->value, state->class_name(dclass), state->type_name(type), state->addr2char(address, &a));
          ret = ANSWER_SATISFIED;
        }
        state->free_rr(&rr);
      }
    }
  }
  /* todo: type == TYPE_AAAA */

  return ret;
}

module_t unnamed_module = { "country", NULL, init, answer, NULL, term };
